@@ -28,6 +28,7 @@
 
 #include "peas-i18n.h"
 #include "peas-engine.h"
+#include "peas-engine-priv.h"
 #include "peas-plugin-info-priv.h"
 #include "peas-plugin-loader.h"
 #include "peas-object-module.h"
@@ -57,6 +58,8 @@
  **/
 G_DEFINE_TYPE (PeasEngine, peas_engine, G_TYPE_OBJECT);
 
+static PeasEngine *default_engine = NULL;
+
 /* Signals */
 enum {
   LOAD_PLUGIN,
@@ -69,8 +72,6 @@ static guint signals[LAST_SIGNAL];
 /* Properties */
 enum {
   PROP_0,
-  PROP_APP_NAME,
-  PROP_SEARCH_PATHS,
   PROP_PLUGIN_LIST,
   PROP_LOADED_PLUGINS
 };
@@ -82,12 +83,17 @@ struct _LoaderInfo {
   PeasObjectModule *module;
 };
 
+typedef struct _SearchPath {
+  gchar *module_dir;
+  gchar *data_dir;
+} SearchPath;
+
 struct _PeasEnginePrivate {
-  gchar *app_name;
-  gchar **search_paths;
+  GList *search_paths;
 
   GList *plugin_list;
-  GHashTable *loaders;
+
+  guint in_dispose : 1;
 };
 
 static void peas_engine_load_plugin_real   (PeasEngine     *engine,
@@ -95,6 +101,9 @@ static void peas_engine_load_plugin_real   (PeasEngine     *engine,
 static void peas_engine_unload_plugin_real (PeasEngine     *engine,
                                             PeasPluginInfo *info);
 
+static gboolean shutdown = FALSE;
+static GHashTable *loaders = NULL;
+
 static void
 load_plugin_info (PeasEngine  *engine,
                   const gchar *filename,
@@ -105,12 +114,14 @@ load_plugin_info (PeasEngine  *engine,
   const gchar *module_name;
 
   info = _peas_plugin_info_new (filename,
-                                engine->priv->app_name,
                                 module_dir,
                                 data_dir);
 
   if (info == NULL)
-    return;
+    {
+      g_warning ("Error loading '%s'", filename);
+      return;
+    }
 
   /* If a plugin with this name has already been loaded
    * drop this one (user plugins override system plugins) */
@@ -118,12 +129,16 @@ load_plugin_info (PeasEngine  *engine,
   if (peas_engine_get_plugin_info (engine, module_name) != NULL)
     _peas_plugin_info_unref (info);
   else
-    engine->priv->plugin_list = g_list_prepend (engine->priv->plugin_list, info);
+    {
+      engine->priv->plugin_list = g_list_prepend (engine->priv->plugin_list,
+                                                  info);
+
+      g_object_notify (G_OBJECT (engine), "plugin-list");
+    }
 }
 
 static void
 load_dir_real (PeasEngine  *engine,
-               const gchar *extension,
                const gchar *module_dir,
                const gchar *data_dir,
                guint        recursions)
@@ -132,7 +147,7 @@ load_dir_real (PeasEngine  *engine,
   GDir *d;
   const gchar *dirent;
 
-  g_debug ("Loading %s/*%s...", module_dir, extension);
+  g_debug ("Loading %s/*.plugin...", module_dir);
 
   d = g_dir_open (module_dir, 0, &error);
 
@@ -147,11 +162,15 @@ load_dir_real (PeasEngine  *engine,
     {
       gchar *filename = g_build_filename (module_dir, dirent, NULL);
 
-      if (g_str_has_suffix (dirent, extension))
-        load_plugin_info (engine, filename, module_dir, data_dir);
-
-      else if (recursions > 0 && g_file_test (filename, G_FILE_TEST_IS_DIR))
-        load_dir_real (engine, extension, filename, data_dir, recursions - 1);
+      if (g_file_test (filename, G_FILE_TEST_IS_DIR))
+        {
+          if (recursions > 0)
+            load_dir_real (engine, filename, data_dir, recursions - 1);
+        }
+      else if (g_str_has_suffix (dirent, ".plugin"))
+        {
+          load_plugin_info (engine, filename, module_dir, data_dir);
+        }
 
       g_free (filename);
     }
@@ -165,38 +184,79 @@ load_dir_real (PeasEngine  *engine,
  *
  * Rescan all the registered directories to find new or updated plugins.
  *
- * Calling this function will make the newly installed plugin infos to be
- * loaded by the engine, so the new plugins can actually be used without
+ * Calling this function will make the newly installed plugin infos
+ * be loaded by the engine, so the new plugins can be used without
  * restarting the application.
  */
 void
 peas_engine_rescan_plugins (PeasEngine *engine)
 {
-  gchar **sp;
-  gchar *extension;
-  guint i;
+  GList *item;
 
   g_return_if_fail (PEAS_IS_ENGINE (engine));
 
-  sp = engine->priv->search_paths;
-  if (sp == NULL)
+  if (engine->priv->search_paths == NULL)
     {
       g_debug ("No search paths where provided");
       return;
     }
 
-  /* Compute the extension of the plugin files. */
-  extension = g_strdup_printf (".%s-plugin", engine->priv->app_name);
-  for (i = 0; extension[i] != '\0'; ++i)
-    extension[i] = g_ascii_tolower (extension[i]);
+  g_object_freeze_notify (G_OBJECT (engine));
 
   /* Go and read everything from the provided search paths */
-  for (i = 0; sp[i] != NULL; i += 2)
-    load_dir_real (engine, extension, sp[i], sp[i + 1], 1);
+  for (item = engine->priv->search_paths; item != NULL; item = item->next)
+    {
+      SearchPath *sp = (SearchPath *) item->data;
+      load_dir_real (engine, sp->module_dir, sp->data_dir, 1);
+    }
 
-  g_object_notify (G_OBJECT (engine), "plugin-list");
+  g_object_thaw_notify (G_OBJECT (engine));
+}
 
-  g_free (extension);
+/**
+ * peas_engine_add_search_path:
+ * @engine: A #PeasEngine.
+ * @module_dir: the plugin module directory.
+ * @data_dir: (allow-none): the plugin data directory.
+ *
+ * This function appends a search path to the list of paths where to
+ * look for plugins.
+ *
+ * A so-called "search path" actually consists of both a
+ * module directory (where the shared libraries or language modules
+ * lie) and a data directory (where the plugin data is).
+ *
+ * The plugin will be able to use a correct data dir depending on
+ * where it is installed, hence allowing to keep the plugin agnostic
+ * when it comes to installation location: the same plugin can be
+ * installed either in the system path or in the user's home directory,
+ * without taking other special care than using
+ * peas_plugin_info_get_data_dir() when looking for its data files.
+ *
+ * If @data_dir is %NULL, then it is set to the same value as
+ * @module_dir.
+ */
+void
+peas_engine_add_search_path (PeasEngine *engine,
+                             const gchar *module_dir,
+                             const gchar *data_dir)
+{
+  SearchPath *sp;
+
+  g_return_if_fail (PEAS_IS_ENGINE (engine));
+  g_return_if_fail (module_dir != NULL);
+
+  sp = g_slice_new (SearchPath);
+  sp->module_dir = g_strdup (module_dir);
+  sp->data_dir = g_strdup (data_dir ? data_dir : module_dir);
+
+  /* Appending to a list is bad, but is easier to handle wrt refreshing
+   * the plugin list. */
+  engine->priv->search_paths = g_list_append (engine->priv->search_paths, sp);
+
+  g_object_freeze_notify (G_OBJECT (engine));
+  load_dir_real (engine, sp->module_dir, sp->data_dir, 1);
+  g_object_thaw_notify (G_OBJECT (engine));
 }
 
 static guint
@@ -231,53 +291,23 @@ loader_destroy (LoaderInfo *info)
   g_free (info);
 }
 
-static LoaderInfo *
-add_loader (PeasEngine       *engine,
-            const gchar      *loader_id,
-            PeasObjectModule *module,
-            PeasPluginLoader *loader)
-{
-  LoaderInfo *info;
-
-  info = g_new (LoaderInfo, 1);
-  info->loader = loader;
-  info->module = module;
-
-  g_hash_table_insert (engine->priv->loaders, g_strdup (loader_id), info);
-  return info;
-}
-
 static void
 peas_engine_init (PeasEngine *engine)
 {
-  if (!g_module_supported ())
-    {
-      g_warning ("libpeas is not able to initialize the plugins engine.");
-      return;
-    }
-
   engine->priv = G_TYPE_INSTANCE_GET_PRIVATE (engine,
                                               PEAS_TYPE_ENGINE,
                                               PeasEnginePrivate);
 
-  /* mapping from loadername -> loader object */
-  engine->priv->loaders = g_hash_table_new_full (hash_lowercase,
-                                                 (GEqualFunc) equal_lowercase,
-                                                 (GDestroyNotify) g_free,
-                                                 (GDestroyNotify) loader_destroy);
-}
+  engine->priv->in_dispose = FALSE;
 
-static void
-peas_engine_constructed (GObject *object)
-{
-  peas_engine_rescan_plugins (PEAS_ENGINE (object));
+  peas_engine_enable_loader (engine, "C");
 }
 
 static void
 loader_garbage_collect (const gchar *id,
                         LoaderInfo  *info)
 {
-  if (info->loader)
+  if (info != NULL && info->loader != NULL)
     peas_plugin_loader_garbage_collect (info->loader);
 }
 
@@ -293,11 +323,46 @@ loader_garbage_collect (const gchar *id,
 void
 peas_engine_garbage_collect (PeasEngine *engine)
 {
-  g_hash_table_foreach (engine->priv->loaders,
+  g_return_if_fail (PEAS_IS_ENGINE (engine));
+
+  g_hash_table_foreach (loaders,
                         (GHFunc) loader_garbage_collect,
                         NULL);
 }
 
+static GObject *
+peas_engine_constructor (GType                  type,
+                         guint                  n_construct_params,
+                         GObjectConstructParam *construct_params)
+{
+  GObject *object;
+
+  /* We don't support calling PeasEngine API without module support */
+  if (!g_module_supported ())
+    {
+      g_error ("libpeas is not able to create the "
+               "plugins engine as modules are not supported.");
+    }
+
+  if (shutdown)
+    {
+      g_error ("libpeas cannot create a plugin engine "
+               "as it has been shutdown.");
+    }
+
+  object = G_OBJECT_CLASS (peas_engine_parent_class)->constructor (type,
+                                                                   n_construct_params,
+                                                                   construct_params);
+
+  if (default_engine == NULL)
+    {
+      default_engine = PEAS_ENGINE (object);
+      g_object_add_weak_pointer (object, (gpointer *) &default_engine);
+    }
+
+  return object;
+}
+
 static void
 peas_engine_set_property (GObject      *object,
                           guint         prop_id,
@@ -308,12 +373,6 @@ peas_engine_set_property (GObject      *object,
 
   switch (prop_id)
     {
-    case PROP_APP_NAME:
-      engine->priv->app_name = g_value_dup_string (value);
-      break;
-    case PROP_SEARCH_PATHS:
-      engine->priv->search_paths = g_value_dup_boxed (value);
-      break;
     case PROP_LOADED_PLUGINS:
       peas_engine_set_loaded_plugins (engine,
                                       (const gchar **) g_value_get_boxed (value));
@@ -334,12 +393,6 @@ peas_engine_get_property (GObject    *object,
 
   switch (prop_id)
     {
-    case PROP_APP_NAME:
-      g_value_set_string (value, engine->priv->app_name);
-      break;
-    case PROP_SEARCH_PATHS:
-      g_value_set_boxed (value, engine->priv->search_paths);
-      break;
     case PROP_PLUGIN_LIST:
       g_value_set_pointer (value,
                            (gpointer) peas_engine_get_plugin_list (engine));
@@ -355,30 +408,47 @@ peas_engine_get_property (GObject    *object,
 }
 
 static void
-peas_engine_finalize (GObject *object)
+peas_engine_dispose (GObject *object)
 {
   PeasEngine *engine = PEAS_ENGINE (object);
   GList *item;
 
+  engine->priv->in_dispose = TRUE;
+
   /* First unload all the plugins */
   for (item = engine->priv->plugin_list; item; item = item->next)
     {
       PeasPluginInfo *info = PEAS_PLUGIN_INFO (item->data);
 
       if (peas_plugin_info_is_loaded (info))
-        peas_engine_unload_plugin_real (engine, info);
+        peas_engine_unload_plugin (engine, info);
     }
 
-  /* unref the loaders */
-  g_hash_table_destroy (engine->priv->loaders);
+  G_OBJECT_CLASS (peas_engine_parent_class)->dispose (object);
+}
 
-  /* and finally free the infos */
+static void
+peas_engine_finalize (GObject *object)
+{
+  PeasEngine *engine = PEAS_ENGINE (object);
+  GList *item;
+
+  /* free the infos */
   for (item = engine->priv->plugin_list; item; item = item->next)
     _peas_plugin_info_unref (PEAS_PLUGIN_INFO (item->data));
 
+  /* free the search path list */
+  for (item = engine->priv->search_paths; item; item = item->next)
+    {
+      SearchPath *sp = (SearchPath *) item->data;
+
+      g_free (sp->module_dir);
+      g_free (sp->data_dir);
+      g_slice_free (SearchPath, sp);
+    }
+
   g_list_free (engine->priv->plugin_list);
-  g_strfreev (engine->priv->search_paths);
-  g_free (engine->priv->app_name);
+  g_list_free (engine->priv->search_paths);
 
   G_OBJECT_CLASS (peas_engine_parent_class)->finalize (object);
 }
@@ -389,76 +459,15 @@ peas_engine_class_init (PeasEngineClass *klass)
   GType the_type = G_TYPE_FROM_CLASS (klass);
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
+  object_class->constructor = peas_engine_constructor;
   object_class->set_property = peas_engine_set_property;
   object_class->get_property = peas_engine_get_property;
-  object_class->constructed = peas_engine_constructed;
+  object_class->dispose = peas_engine_dispose;
   object_class->finalize = peas_engine_finalize;
 
   klass->load_plugin = peas_engine_load_plugin_real;
   klass->unload_plugin = peas_engine_unload_plugin_real;
 
-  /**
-   * PeasEngine:app-name:
-   *
-   * The application name. Filename extension and section header for
-   * #PeasPluginInfo files are actually derived from this value.
-   *
-   * For instance, if app-name is "Gedit", then info files will have
-   * the .gedit-plugin extension, and the engine will look for a
-   * "Gedit Plugin" section in it to load the plugin data.
-   */
-  g_object_class_install_property (object_class,
-                                   PROP_APP_NAME,
-                                   g_param_spec_string ("app-name",
-                                                        "Application Name",
-                                                        "The application name",
-                                                        "Peas",
-                                                        G_PARAM_READWRITE |
-                                                        G_PARAM_CONSTRUCT_ONLY |
-                                                        G_PARAM_STATIC_STRINGS));
-
-  /**
-   * PeasEngine:search-paths:
-   *
-   * The list of path where to look for plugins.
-   *
-   * A so-called "search path" actually consists on a couple of both a
-   * module directory (where the shared libraries or language modules
-   * lie) and a data directory (where the plugin data is).
-   *
-   * The #PeasPlugin will be able to use a correct data dir depending on
-   * where it is installed, hence allowing to keep the plugin agnostic
-   * when it comes to installation location: the same plugin can be
-   * installed both in the system path or in the user's home directory,
-   * without taking other special care than using
-   * peas_plugin_info_get_data_dir() when looking for its data files.
-   *
-   * Concretely, this property contains a %NULL-terminated array of
-   * strings, whose even-indexed strings are module directories, and
-   * odd-indexed ones are the associated data directories.  Here is an
-   * example of such an array:
-   * |[
-   * gchar const * const search_paths[] = {
-   *         // Plugins in ~/.config
-   *         g_build_filename (g_get_user_config_dir (), "example/plugins", NULL),
-   *         g_build_filename (g_get_user_config_dir (), "example/plugins", NULL),
-   *         // System plugins
-   *         EXAMPLE_PREFIX "/lib/example/plugins/",
-   *         EXAMPLE_PREFIX "/share/example/plugins/",
-   *         NULL
-   * };
-   * ]|
-   */
-  g_object_class_install_property (object_class,
-                                   PROP_SEARCH_PATHS,
-                                   g_param_spec_boxed ("search-paths",
-                                                       "Search paths",
-                                                       "The list of paths where to look for plugins",
-                                                       G_TYPE_STRV,
-                                                       G_PARAM_READWRITE |
-                                                       G_PARAM_CONSTRUCT_ONLY |
-                                                       G_PARAM_STATIC_STRINGS));
-
   /**
    * PeasEngine:plugin-list:
    *
@@ -494,6 +503,8 @@ peas_engine_class_init (PeasEngineClass *klass)
    *                    "loaded-plugins",
    *                    G_SETTINGS_BIND_DEFAULT);
    * ]|
+   *
+   * Note: notify will not be called when the engine is being destroyed.
    */
   g_object_class_install_property (object_class,
                                    PROP_LOADED_PLUGINS,
@@ -558,6 +569,12 @@ peas_engine_class_init (PeasEngineClass *klass)
   /* We are doing some global initialization here as there is currently no
    * global init function for libpeas. */
   peas_debug_init ();
+
+  /* mapping from loadername -> loader object */
+  loaders = g_hash_table_new_full (hash_lowercase,
+                                   (GEqualFunc) equal_lowercase,
+                                   (GDestroyNotify) g_free,
+                                   (GDestroyNotify) loader_destroy);
 }
 
 static PeasObjectModule *
@@ -599,101 +616,91 @@ try_to_open_loader_module (const gchar *loader_id,
   return module;
 }
 
-static LoaderInfo *
-load_plugin_loader (PeasEngine  *engine,
-                    const gchar *loader_id)
-{
-  guint i;
-  gchar *lc_loader_id;
-  PeasObjectModule *module;
-  PeasPluginLoader *loader;
-
-  /* We need to ensure we use the lowercase loader_id */
-  lc_loader_id = g_strdup (loader_id);
-  for (i = 0; lc_loader_id[i] != '\0'; ++i)
-    lc_loader_id[i] = g_ascii_tolower (lc_loader_id[i]);
-
-  module = try_to_open_loader_module (lc_loader_id, FALSE);
-  if (module == NULL)
-    module = try_to_open_loader_module (lc_loader_id, TRUE);
-
-  g_free (lc_loader_id);
-
-  if (module == NULL)
-    {
-      g_warning ("Loader '%s' could not be loaded", loader_id);
-      return add_loader (engine, loader_id, NULL, NULL);
-    }
-
-  loader = (PeasPluginLoader *) peas_object_module_create_object (module, PEAS_TYPE_PLUGIN_LOADER, 0, NULL);
-
-  if (loader == NULL || !PEAS_IS_PLUGIN_LOADER (loader))
-    {
-      g_warning ("Loader '%s' is not a valid PeasPluginLoader instance", loader_id);
-      if (loader != NULL && G_IS_OBJECT (loader))
-        g_object_unref (loader);
-      module = NULL;
-      loader = NULL;
-    }
-
-  g_type_module_unuse (G_TYPE_MODULE (module));
-
-  return add_loader (engine, loader_id, module, loader);
-}
-
 static PeasPluginLoader *
 get_plugin_loader (PeasEngine     *engine,
                    PeasPluginInfo *info)
 {
-  const gchar *loader_id;
   LoaderInfo *loader_info;
+  gchar *loader_id;
 
-  loader_id = info->loader;
-
-  loader_info = (LoaderInfo *) g_hash_table_lookup (engine->priv->loaders,
-                                                    loader_id);
+  loader_info = (LoaderInfo *) g_hash_table_lookup (loaders, info->loader);
 
+  /* The loader has not been enabled. */
   if (loader_info == NULL)
-    loader_info = load_plugin_loader (engine, loader_id);
+    return NULL;
+
+  /* The loader has already been loaded. */
+  if (loader_info->loader != NULL)
+    return loader_info->loader;
+
+  /* We need to ensure we use the lowercase loader_id */
+  loader_id = g_ascii_strdown (info->loader, -1);
+
+  loader_info->module = try_to_open_loader_module (loader_id, FALSE);
+  if (loader_info->module == NULL)
+    loader_info->module = try_to_open_loader_module (loader_id, TRUE);
+
+  g_free (loader_id);
+
+  if (loader_info->module == NULL)
+    {
+      g_hash_table_insert (loaders, info->loader, NULL);
+      return NULL;
+    }
+
+  loader_info->loader = PEAS_PLUGIN_LOADER (
+        peas_object_module_create_object (loader_info->module,
+                                          PEAS_TYPE_PLUGIN_LOADER,
+                                          0, NULL));
+
+  g_type_module_unuse (G_TYPE_MODULE (loader_info->module));
+
+  if (loader_info->loader == NULL ||
+      !PEAS_IS_PLUGIN_LOADER (loader_info->loader))
+    {
+      g_warning ("Loader '%s' is not a valid PeasPluginLoader instance",
+                 info->loader);
+
+      if (G_IS_OBJECT (loader_info->loader))
+        g_object_unref (loader_info->loader);
+
+      g_object_unref (loader_info->module);
+      g_hash_table_insert (loaders, info->loader, NULL);
+      return NULL;
+    }
 
   return loader_info->loader;
 }
 
 /**
- * peas_engine_disable_loader:
+ * peas_engine_enable_loader:
  * @engine: A #PeasEngine.
- * @loader_id: The id of the loader to inhibit.
+ * @loader_id: The id of the loader to enable.
  *
- * Disable a loader, avoiding using the plugins written using the
- * related language to be loaded. This method must be called before
- * any plugin relying on the loader @loader_id has been loaded.
+ * Enable a loader, enables a loader for plugins.
+ * The C plugin loader is always enabled.
  *
- * For instance, the following code will prevent any python plugin
+ * For instance, the following code will enable python plugin
  * from being loaded:
  * |[
- * peas_engine_disable_loader (engine, "python");
+ * peas_engine_enable_loader (engine, "python");
  * ]|
+ *
+ * Note: plugin loaders are shared across #PeasEngines so enabling
+ *       a loader on one #PeasEngine will enable it on all #PeasEngines.
  **/
 void
-peas_engine_disable_loader (PeasEngine  *engine,
-                            const gchar *loader_id)
+peas_engine_enable_loader (PeasEngine  *engine,
+                           const gchar *loader_id)
 {
-  LoaderInfo *loader_info;
-
   g_return_if_fail (PEAS_IS_ENGINE (engine));
+  g_return_if_fail (loader_id != NULL && *loader_id != '\0');
 
-  loader_info = (LoaderInfo *) g_hash_table_lookup (engine->priv->loaders,
-                                                    loader_id);
-  if (loader_info != NULL)
-    {
-      g_warning ("Loader '%s' cannot be disabled as it is already loaded",
-                 loader_id);
-      return;
-    }
+  if (g_hash_table_lookup_extended (loaders, loader_id, NULL, NULL))
+    return;
 
-  /* By adding a NULL loader, we simulate a failed load attempt, effectively
-   * disabling the loader for further use. */
-  add_loader (engine, loader_id, NULL, NULL);
+  /* The loader is loaded in get_plugin_loader() */
+  g_hash_table_insert (loaders, g_strdup (loader_id), g_new0 (LoaderInfo, 1));
 }
 
 /**
@@ -709,6 +716,8 @@ peas_engine_disable_loader (PeasEngine  *engine,
 const GList *
 peas_engine_get_plugin_list (PeasEngine *engine)
 {
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
+
   return engine->priv->plugin_list;
 }
 
@@ -724,15 +733,21 @@ compare_plugin_info_and_name (PeasPluginInfo *info,
  * @engine: A #PeasEngine.
  * @plugin_name: A plugin name.
  *
- * Returns: the #PeasPluginInfo corresponding with a given plugin name.
+ * Returns: the #PeasPluginInfo corresponding with
+ * a given plugin module name.
  */
 PeasPluginInfo *
 peas_engine_get_plugin_info (PeasEngine  *engine,
                              const gchar *plugin_name)
 {
-  GList *l = g_list_find_custom (engine->priv->plugin_list,
-                                 plugin_name,
-                                 (GCompareFunc) compare_plugin_info_and_name);
+  GList *l;
+
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (plugin_name != NULL, NULL);
+
+  l = g_list_find_custom (engine->priv->plugin_list,
+                          plugin_name,
+                          (GCompareFunc) compare_plugin_info_and_name);
 
   return l == NULL ? NULL : (PeasPluginInfo *) l->data;
 }
@@ -749,7 +764,7 @@ load_plugin (PeasEngine     *engine,
   if (peas_plugin_info_is_loaded (info))
     return TRUE;
 
-  if (!peas_plugin_info_is_available (info))
+  if (!peas_plugin_info_is_available (info, NULL))
     return FALSE;
 
   /* We set the plugin info as loaded before trying to load the dependencies,
@@ -763,11 +778,23 @@ load_plugin (PeasEngine     *engine,
       if (!dep_info)
         {
           g_warning ("Plugin not found: %s", dependencies[i]);
-          return FALSE;
+          g_set_error (&info->error,
+                       PEAS_PLUGIN_INFO_ERROR,
+                       PEAS_PLUGIN_INFO_ERROR_DEP_NOT_FOUND,
+                       _("Dependency '%s' was not found"),
+                       dependencies[i]);
+          goto error;
         }
 
       if (!peas_engine_load_plugin (engine, dep_info))
-        return FALSE;
+        {
+          g_set_error (&info->error,
+                       PEAS_PLUGIN_INFO_ERROR,
+                       PEAS_PLUGIN_INFO_ERROR_LOADING_FAILED,
+                       _("Dependency '%s' failed to load"),
+                       dep_info->name);
+          goto error;
+        }
     }
 
   loader = get_plugin_loader (engine, info);
@@ -776,19 +803,32 @@ load_plugin (PeasEngine     *engine,
     {
       g_warning ("Could not find loader '%s' for plugin '%s'",
                  info->loader, info->name);
-      info->available = FALSE;
-      return FALSE;
+      g_set_error (&info->error,
+                   PEAS_PLUGIN_INFO_ERROR,
+                   PEAS_PLUGIN_INFO_ERROR_LOADER_NOT_FOUND,
+                   _("Plugin loader '%s' was not found"),
+                   info->loader);
+      goto error;
     }
 
   if (!peas_plugin_loader_load (loader, info))
     {
       g_warning ("Error loading plugin '%s'", info->name);
-      info->loaded = FALSE;
-      info->available = FALSE;
-      return FALSE;
+      g_set_error (&info->error,
+                   PEAS_PLUGIN_INFO_ERROR,
+                   PEAS_PLUGIN_INFO_ERROR_LOADING_FAILED,
+                   _("Failed to load"));
+      goto error;
     }
 
   return TRUE;
+
+error:
+
+  info->loaded = FALSE;
+  info->available = FALSE;
+
+  return FALSE;
 }
 
 static void
@@ -814,9 +854,10 @@ gboolean
 peas_engine_load_plugin (PeasEngine     *engine,
                          PeasPluginInfo *info)
 {
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), FALSE);
   g_return_val_if_fail (info != NULL, FALSE);
 
-  if (!peas_plugin_info_is_available (info))
+  if (!peas_plugin_info_is_available (info, NULL))
     return FALSE;
 
   if (peas_plugin_info_is_loaded (info))
@@ -836,7 +877,7 @@ peas_engine_unload_plugin_real (PeasEngine     *engine,
   PeasPluginLoader *loader;
 
   if (!peas_plugin_info_is_loaded (info) ||
-      !peas_plugin_info_is_available (info))
+      !peas_plugin_info_is_available (info, NULL))
     return;
 
   /* We set the plugin info as unloaded before trying to unload the
@@ -862,7 +903,8 @@ peas_engine_unload_plugin_real (PeasEngine     *engine,
   peas_plugin_loader_garbage_collect (loader);
   peas_plugin_loader_unload (loader, info);
 
-  g_object_notify (G_OBJECT (engine), "loaded-plugins");
+  if (!engine->priv->in_dispose)
+    g_object_notify (G_OBJECT (engine), "loaded-plugins");
 }
 
 /**
@@ -880,6 +922,7 @@ gboolean
 peas_engine_unload_plugin (PeasEngine     *engine,
                            PeasPluginInfo *info)
 {
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), FALSE);
   g_return_val_if_fail (info != NULL, FALSE);
 
   if (!peas_plugin_info_is_loaded (info))
@@ -908,12 +951,13 @@ peas_engine_provides_extension (PeasEngine     *engine,
 }
 
 /**
- * peas_engine_create_extensionv
+ * peas_engine_create_extensionv:
  * @engine: A #PeasEngine.
  * @info: A loaded #PeasPluginInfo.
  * @extension_type: The implemented extension #GType.
  * @n_parameters: the length of the @parameters array.
- * @parameters: an array of #GParameter.
+ * @parameters: (allow-none) (array length=n_parameters):
+ *   an array of #GParameter.
  *
  * If the plugin identified by @info implements the @extension_type interface,
  * then this function will return a new instance of this implementation,
@@ -921,8 +965,10 @@ peas_engine_provides_extension (PeasEngine     *engine,
  *
  * See peas_engine_create_extension() for more information.
  *
- * Returns: a new instance of #PeasExtension wrapping the @extension_type
- * instance, or %NULL.
+ * Returns: (transfer full): a new instance of #PeasExtension wrapping
+ * the @extension_type instance, or %NULL.
+ *
+ * Rename to: peas_engine_create_extension
  */
 PeasExtension *
 peas_engine_create_extensionv (PeasEngine     *engine,
@@ -935,6 +981,7 @@ peas_engine_create_extensionv (PeasEngine     *engine,
 
   g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
   g_return_val_if_fail (info != NULL, NULL);
+  g_return_val_if_fail (peas_plugin_info_is_loaded (info), NULL);
 
   loader = get_plugin_loader (engine, info);
   return peas_plugin_loader_create_extension (loader, info, extension_type,
@@ -942,7 +989,7 @@ peas_engine_create_extensionv (PeasEngine     *engine,
 }
 
 /**
- * peas_engine_create_extension_valist
+ * peas_engine_create_extension_valist: (skip)
  * @engine: A #PeasEngine.
  * @info: A loaded #PeasPluginInfo.
  * @extension_type: The implemented extension #GType.
@@ -956,8 +1003,8 @@ peas_engine_create_extensionv (PeasEngine     *engine,
  *
  * See peas_engine_create_extension() for more information.
  *
- * Returns: a new instance of #PeasExtension wrapping the @extension_type
- * instance, or %NULL.
+ * Returns: a new instance of #PeasExtension wrapping
+ * the @extension_type instance, or %NULL.
  */
 PeasExtension *
 peas_engine_create_extension_valist (PeasEngine     *engine,
@@ -992,7 +1039,7 @@ peas_engine_create_extension_valist (PeasEngine     *engine,
 }
 
 /**
- * peas_engine_create_extension:
+ * peas_engine_create_extension: (skip)
  * @engine: A #PeasEngine.
  * @info: A loaded #PeasPluginInfo.
  * @extension_type: The implemented extension #GType.
@@ -1013,8 +1060,8 @@ peas_engine_create_extension_valist (PeasEngine     *engine,
  * principle of never giving you the actual object (also because it might as
  * well *not* be an actual object).
  *
- * Returns: a new instance of #PeasExtension wrapping the @extension_type
- * instance, or %NULL.
+ * Returns: a new instance of #PeasExtension wrapping
+ * the @extension_type instance, or %NULL.
  */
 PeasExtension *
 peas_engine_create_extension (PeasEngine     *engine,
@@ -1044,14 +1091,19 @@ peas_engine_create_extension (PeasEngine     *engine,
  * Please note that the returned array is a newly allocated one: you will need
  * to free it using g_strfreev().
  *
- * Returns: A newly-allocated NULL-terminated array of strings, or %NULL.
+ * Returns: (transfer full) (array zero-terminated=1): A newly-allocated
+ * %NULL-terminated array of strings.
  */
 gchar **
 peas_engine_get_loaded_plugins (PeasEngine *engine)
 {
-  GArray *array = g_array_new (TRUE, FALSE, sizeof (gchar *));
+  GArray *array;
   GList *pl;
 
+  g_return_val_if_fail (PEAS_IS_ENGINE (engine), NULL);
+
+  array = g_array_new (TRUE, FALSE, sizeof (gchar *));
+
   for (pl = engine->priv->plugin_list; pl; pl = pl->next)
     {
       PeasPluginInfo *info = (PeasPluginInfo *) pl->data;
@@ -1072,6 +1124,10 @@ string_in_strv (const gchar  *needle,
                 const gchar **haystack)
 {
   guint i;
+
+  if (haystack == NULL)
+    return FALSE;
+
   for (i = 0; haystack[i] != NULL; i++)
     if (strcmp (haystack[i], needle) == 0)
       return TRUE;
@@ -1081,11 +1137,14 @@ string_in_strv (const gchar  *needle,
 /**
  * peas_engine_set_loaded_plugins:
  * @engine: A #PeasEngine.
- * @plugin_names: A %NULL-terminated array of plugin names.
+ * @plugin_names: (allow-none) (array zero-terminated=1): A %NULL-terminated
+ *  array of plugin names, or %NULL.
  *
  * Sets the list of loaded plugins for @engine. When this function is called,
  * the #PeasEngine will load all the plugins whose names are in @plugin_names,
  * and ensures all other active plugins are unloaded.
+ *
+ * If @plugin_names is %NULL, all plugins will be unloaded.
  */
 void
 peas_engine_set_loaded_plugins (PeasEngine   *engine,
@@ -1093,6 +1152,8 @@ peas_engine_set_loaded_plugins (PeasEngine   *engine,
 {
   GList *pl;
 
+  g_return_if_fail (PEAS_IS_ENGINE (engine));
+
   for (pl = engine->priv->plugin_list; pl; pl = pl->next)
     {
       PeasPluginInfo *info = (PeasPluginInfo *) pl->data;
@@ -1100,7 +1161,7 @@ peas_engine_set_loaded_plugins (PeasEngine   *engine,
       gboolean is_loaded;
       gboolean to_load;
 
-      if (!peas_plugin_info_is_available (info))
+      if (!peas_plugin_info_is_available (info, NULL))
         continue;
 
       module_name = peas_plugin_info_get_module_name (info);
@@ -1117,22 +1178,56 @@ peas_engine_set_loaded_plugins (PeasEngine   *engine,
 
 /**
  * peas_engine_new:
- * @app_name: (allow-none): The name of the application
- * @search_paths: The paths where to look for plugins
  *
- * Returns a new #PeasEngine object.
- * See the properties description for more information about the parameters.
+ * Return a new instance of #PeasEngine.
+ * If no default #PeasEngine has been instantiated yet,
+ * the first call of this function will set the default
+ * engine as the new instance of #PeasEngine.
  *
- * Returns: a newly created #PeasEngine object.
+ * Returns: a new instance of #PeasEngine.
  */
 PeasEngine *
-peas_engine_new (const gchar  *app_name,
-                 const gchar **search_paths)
+peas_engine_new (void)
 {
-  g_return_val_if_fail (search_paths != NULL, NULL);
+  return PEAS_ENGINE (g_object_new (PEAS_TYPE_ENGINE, NULL));
+}
 
-  return PEAS_ENGINE (g_object_new (PEAS_TYPE_ENGINE,
-                                    "app-name", app_name,
-                                    "search-paths", search_paths,
-                                    NULL));
+/**
+ * peas_engine_get_default:
+ *
+ * Return the existing instance of #PeasEngine or a subclass of it.
+ * If no #PeasEngine subclass has been instantiated yet, the first call
+ * of this function will return a new instance of #PeasEngine.
+ *
+ * Returns: (transfer none): the existing instance of #PeasEngine.
+ */
+PeasEngine *
+peas_engine_get_default (void)
+{
+  /* peas_engine_new() will cause the default to be set for us */
+  if (default_engine == NULL)
+    return peas_engine_new ();
+
+  return default_engine;
+}
+
+/*
+ * peas_engine_shutdown:
+ *
+ * Frees memory shared by PeasEngines.
+ * No libpeas API should be called after calling this.
+ */
+void
+peas_engine_shutdown (void)
+{
+  if (shutdown)
+    return;
+
+  shutdown = TRUE;
+
+  if (loaders != NULL)
+    {
+      g_hash_table_destroy (loaders);
+      loaders = NULL;
+    }
 }
