@@ -53,28 +53,17 @@
  * static void
  * on_extension_added (PeasExtensionSet *set,
  *                     PeasPluginInfo   *info,
- *                     PeasExtension    *exten,
- *                     GtkWindow        *window)
+ *                     PeasActivatable  *activatable)
  * {
- *   peas_extension_call (exten, "activate", window);
+ *   peas_activatable_activate (activatable);
  * }
  *
  * static void
  * on_extension_removed (PeasExtensionSet *set,
  *                       PeasPluginInfo   *info,
- *                       PeasExtension    *exten,
- *                       GtkWindow        *window)
+ *                       PeasActivatable  *activatable)
  * {
- *   peas_extension_call (exten, "deactivate", window);
- * }
- *
- * static gboolean
- * on_delete_event (GtkWidget        *window,
- *                  GdkEvent         *event,
- *                  PeasExtensionSet *set)
- * {
- *   peas_extension_set_call (set, "deactivate");
- *   return FALSE;
+ *   peas_activatable_deactivate (activatable);
  * }
  *
  * PeasExtensionSet *
@@ -87,11 +76,9 @@
  *                                 "object", window, NULL);
  *   peas_extension_set_call (set, "activate");
  *   g_signal_connect (set, "extension-added",
- *                     G_CALLBACK (on_extension_added), window);
+ *                     G_CALLBACK (on_extension_added), NULL);
  *   g_signal_connect (set, "extension-removed",
- *                     G_CALLBACK (on_extension_removed), window);
- *   g_signal_connect (window, "delete-event",
- *                     G_CALLBACK (on_delete_event), set);
+ *                     G_CALLBACK (on_extension_removed), NULL);
  *   return set;
  * }
  * ]|
@@ -116,6 +103,11 @@ typedef struct {
   PeasExtension *exten;
 } ExtensionItem;
 
+typedef struct {
+  guint n_parameters;
+  GParameter *parameters;
+} PeasParameterArray;
+
 /* Signals */
 enum {
   EXTENSION_ADDED,
@@ -162,7 +154,6 @@ peas_extension_set_set_property (GObject      *object,
     {
     case PROP_ENGINE:
       set->priv->engine = g_value_get_object (value);
-      g_object_ref (set->priv->engine);
       break;
     case PROP_EXTENSION_TYPE:
       set->priv->exten_type = g_value_get_gtype (value);
@@ -265,9 +256,13 @@ static void
 peas_extension_set_constructed (GObject *object)
 {
   PeasExtensionSet *set = PEAS_EXTENSION_SET (object);
-
   GList *plugins, *l;
 
+  if (set->priv->engine == NULL)
+    set->priv->engine = peas_engine_get_default ();
+
+  g_object_ref (set->priv->engine);
+
   plugins = (GList *) peas_engine_get_plugin_list (set->priv->engine);
   for (l = plugins; l; l = l->next)
     add_extension (set, (PeasPluginInfo *) l->data);
@@ -283,13 +278,22 @@ peas_extension_set_constructed (GObject *object)
 }
 
 static void
-peas_extension_set_finalize (GObject *object)
+peas_extension_set_dispose (GObject *object)
 {
   PeasExtensionSet *set = PEAS_EXTENSION_SET (object);
   GList *l;
 
-  g_signal_handler_disconnect (set->priv->engine, set->priv->load_handler_id);
-  g_signal_handler_disconnect (set->priv->engine, set->priv->unload_handler_id);
+  if (set->priv->load_handler_id != 0)
+    {
+      g_signal_handler_disconnect (set->priv->engine, set->priv->load_handler_id);
+      set->priv->load_handler_id = 0;
+    }
+
+  if (set->priv->unload_handler_id != 0)
+    {
+      g_signal_handler_disconnect (set->priv->engine, set->priv->unload_handler_id);
+      set->priv->unload_handler_id = 0;
+    }
 
   for (l = set->priv->extensions; l;)
     {
@@ -301,10 +305,16 @@ peas_extension_set_finalize (GObject *object)
     {
       while (set->priv->n_parameters-- > 0)
         g_value_unset (&set->priv->parameters[set->priv->n_parameters].value);
+
       g_free (set->priv->parameters);
+      set->priv->parameters = NULL;
     }
 
-  g_object_unref (set->priv->engine);
+  if (set->priv->engine != NULL)
+    {
+      g_object_unref (set->priv->engine);
+      set->priv->engine = NULL;
+    }
 }
 
 static gboolean
@@ -334,7 +344,7 @@ peas_extension_set_class_init (PeasExtensionSetClass *klass)
   object_class->set_property = peas_extension_set_set_property;
   object_class->get_property = peas_extension_set_get_property;
   object_class->constructed = peas_extension_set_constructed;
-  object_class->finalize = peas_extension_set_finalize;
+  object_class->dispose = peas_extension_set_dispose;
 
   klass->call = peas_extension_set_call_real;
 
@@ -371,9 +381,10 @@ peas_extension_set_class_init (PeasExtensionSetClass *klass)
    * @info: A #PeasPluginInfo.
    * @exten: A #PeasExtension.
    *
-   * The extension-added signal is emitted when a new extension is about to be
+   * The extension-removed signal is emitted when a new extension is about to be
    * removed from the #PeasExtensionSet. It happens when a plugin implementing
-   * the extension set's extension type is unloaded.
+   * the extension set's extension type is unloaded, or when the
+   * #PeasExtensionSet itself is destroyed.
    *
    * You should connect to this signal in order to clean up the extensions
    * when their plugin is unload. Note that this signal is not fired for the
@@ -452,7 +463,7 @@ peas_extension_set_get_extension (PeasExtensionSet *set,
 }
 
 /**
- * peas_extension_set_call:
+ * peas_extension_set_call: (skip)
  * @set: A #PeasExtensionSet.
  * @method_name: the name of the method that should be called.
  * @Varargs: arguments for the method.
@@ -471,6 +482,9 @@ peas_extension_set_call (PeasExtensionSet *set,
   va_list args;
   gboolean result;
 
+  g_return_val_if_fail (PEAS_IS_EXTENSION_SET (set), FALSE);
+  g_return_val_if_fail (method_name != NULL, FALSE);
+
   va_start (args, method_name);
   result = peas_extension_set_call_valist (set, method_name, args);
   va_end (args);
@@ -479,7 +493,7 @@ peas_extension_set_call (PeasExtensionSet *set,
 }
 
 /**
- * peas_extension_set_call_valist:
+ * peas_extension_set_call_valist: (skip)
  * @set: A #PeasExtensionSet.
  * @method_name: the name of the method that should be called.
  * @va_args: the arguments for the method.
@@ -497,14 +511,22 @@ peas_extension_set_call_valist (PeasExtensionSet *set,
 {
   GICallableInfo *callable_info;
   GIArgument *args;
+  gint n_args;
 
   g_return_val_if_fail (PEAS_IS_EXTENSION_SET (set), FALSE);
   g_return_val_if_fail (method_name != NULL, FALSE);
 
   callable_info = peas_gi_get_method_info (set->priv->exten_type, method_name);
-  args = g_newa (GIArgument, g_callable_info_get_n_args (callable_info));
+  g_return_val_if_fail (callable_info != NULL, FALSE);
+
+  n_args = g_callable_info_get_n_args (callable_info);
+  g_return_val_if_fail (n_args >= 0, FALSE);
+
+  args = g_newa (GIArgument, n_args);
   peas_gi_valist_to_arguments (callable_info, va_args, args, NULL);
 
+  g_base_info_unref ((GIBaseInfo *) callable_info);
+
   return peas_extension_set_callv (set, method_name, args);
 }
 
@@ -519,6 +541,8 @@ peas_extension_set_call_valist (PeasExtensionSet *set,
  * See peas_extension_callv() for more information.
  *
  * Return value: %TRUE on successful call.
+ *
+ * Rename to: peas_extension_set_call
  */
 gboolean
 peas_extension_set_callv (PeasExtensionSet *set,
@@ -536,16 +560,20 @@ peas_extension_set_callv (PeasExtensionSet *set,
 
 /**
  * peas_extension_set_newv:
- * @engine: A #PeasEngine.
+ * @engine: (allow-none): A #PeasEngine, or %NULL.
  * @exten_type: the extension #GType.
  * @n_parameters: the length of the @parameters array.
- * @parameters: an array of #GParameter.
+ * @parameters: (array length=n_parameters): an array of #GParameter.
  *
  * Create a new #PeasExtensionSet for the @exten_type extension type.
  *
+ * If @engine is %NULL, then the default engine will be used.
+ *
  * See peas_extension_set_new() for more information.
  *
  * Returns: (transfer full): a new instance of #PeasExtensionSet.
+ *
+ * Rename to: peas_extension_set_new
  */
 PeasExtensionSet *
 peas_extension_set_newv (PeasEngine *engine,
@@ -555,6 +583,9 @@ peas_extension_set_newv (PeasEngine *engine,
 {
   PeasParameterArray construct_properties = { n_parameters, parameters };
 
+  g_return_val_if_fail (engine == NULL || PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (exten_type), NULL);
+
   return PEAS_EXTENSION_SET (g_object_new (PEAS_TYPE_EXTENSION_SET,
                                            "engine", engine,
                                            "extension-type", exten_type,
@@ -563,8 +594,8 @@ peas_extension_set_newv (PeasEngine *engine,
 }
 
 /**
- * peas_extension_set_new_valist:
- * @engine: A #PeasEngine.
+ * peas_extension_set_new_valist: (skip)
+ * @engine: A #PeasEngine, or %NULL.
  * @exten_type: the extension #GType.
  * @first_property: the name of the first property.
  * @var_args: the value of the first property, followed optionally by more
@@ -572,6 +603,8 @@ peas_extension_set_newv (PeasEngine *engine,
  *
  * Create a new #PeasExtensionSet for the @exten_type extension type.
  *
+ * If @engine is %NULL, then the default engine will be used.
+ *
  * See peas_extension_set_new() for more information.
  *
  * Returns: a new instance of #PeasExtensionSet.
@@ -587,6 +620,9 @@ peas_extension_set_new_valist (PeasEngine  *engine,
   guint n_parameters;
   PeasExtensionSet *set;
 
+  g_return_val_if_fail (engine == NULL || PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (exten_type), NULL);
+
   type_struct = _g_type_struct_ref (exten_type);
 
   if (!_valist_to_parameter_list (exten_type, type_struct, first_property,
@@ -608,8 +644,8 @@ peas_extension_set_new_valist (PeasEngine  *engine,
 }
 
 /**
- * peas_extension_set_new:
- * @engine: A #PeasEngine.
+ * peas_extension_set_new: (skip)
+ * @engine: A #PeasEngine, or %NULL.
  * @exten_type: the extension #GType.
  * @first_property: the name of the first property.
  * @Varargs: the value of the first property, followed optionally by more
@@ -624,6 +660,8 @@ peas_extension_set_new_valist (PeasEngine  *engine,
  * The property values passed to peas_extension_set_new() will be used for the
  * construction of new extension instances.
  *
+ * If @engine is %NULL, then the default engine will be used.
+ *
  * See peas_engine_create_extension() for more information.
  *
  * Returns: a new instance of #PeasExtensionSet.
@@ -637,6 +675,9 @@ peas_extension_set_new (PeasEngine  *engine,
   va_list var_args;
   PeasExtensionSet *set;
 
+  g_return_val_if_fail (engine == NULL || PEAS_IS_ENGINE (engine), NULL);
+  g_return_val_if_fail (G_TYPE_IS_INTERFACE (exten_type), NULL);
+
   va_start (var_args, first_property);
   set = peas_extension_set_new_valist (engine, exten_type, first_property, var_args);
   va_end (var_args);
