@@ -37,27 +37,31 @@ peas_extension_c_init (PeasExtensionC *cexten)
 
 static gboolean
 peas_extension_c_call (PeasExtension *exten,
+                       GType          gtype,
                        const gchar   *method_name,
                        GIArgument    *args,
                        GIArgument    *retval)
 {
   PeasExtensionC *cexten = PEAS_EXTENSION_C (exten);
-  GType gtype;
 
-  gtype = peas_extension_get_extension_type (exten);
+  if (gtype == G_TYPE_INVALID)
+    gtype = peas_extension_get_extension_type (exten);
 
   return peas_method_apply (cexten->instance, gtype, method_name, args, retval);
 }
 
 static void
-peas_extension_c_finalize (GObject *object)
+peas_extension_c_dispose (GObject *object)
 {
   PeasExtensionC *cexten = PEAS_EXTENSION_C (object);
   
   if (cexten->instance)
-    g_object_unref (cexten->instance);
+    {
+      g_object_unref (cexten->instance);
+      cexten->instance = NULL;
+    }
 
-  G_OBJECT_CLASS (peas_extension_c_parent_class)->finalize (object);
+  G_OBJECT_CLASS (peas_extension_c_parent_class)->dispose (object);
 }
 
 static void
@@ -66,7 +70,7 @@ peas_extension_c_class_init (PeasExtensionCClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   PeasExtensionClass *extension_class = PEAS_EXTENSION_CLASS (klass);
 
-  object_class->finalize = peas_extension_c_finalize;
+  object_class->dispose = peas_extension_c_dispose;
 
   extension_class->call = peas_extension_c_call;
 }
