@@ -24,11 +24,15 @@
 #endif
 
 #include <seed.h>
+/* FIXME: Seed Bug 624560 */
+SeedValue seed_value_from_gvalue (SeedContext    ctx,
+                                  GValue        *gval,
+                                  SeedException *exception);
 
 #include "peas-plugin-loader-seed.h"
 #include "peas-extension-seed.h"
 
-G_DEFINE_DYNAMIC_TYPE (PeasPluginLoaderSeed, peas_plugin_loader_seed, PEAS_TYPE_PLUGIN_LOADER);
+G_DEFINE_TYPE (PeasPluginLoaderSeed, peas_plugin_loader_seed, PEAS_TYPE_PLUGIN_LOADER);
 
 typedef struct {
   SeedContext context;
@@ -70,7 +74,7 @@ get_script_for_plugin_info (PeasPluginInfo   *info,
   basename = g_strconcat (peas_plugin_info_get_module_name (info), ".js", NULL);
   filename = g_build_filename (peas_plugin_info_get_module_dir (info), basename, NULL);
 
-  g_debug ("Seed script filename is %s", filename);
+  g_debug ("Seed script filename is '%s'", filename);
 
   g_file_get_contents (filename, &script, NULL, NULL);
 
@@ -114,7 +118,7 @@ peas_plugin_loader_seed_load (PeasPluginLoader *loader,
 
       if (seed_value_is_object (context, extensions))
         {
-          sinfo = g_slice_new (SeedInfo);
+          sinfo = (SeedInfo *) g_slice_new (SeedInfo);
           sinfo->context = context;
           sinfo->extensions = extensions;
           seed_context_ref (context);
@@ -136,7 +140,7 @@ peas_plugin_loader_seed_provides_extension  (PeasPluginLoader *loader,
 {
   PeasPluginLoaderSeed *sloader = PEAS_PLUGIN_LOADER_SEED (loader);
   SeedInfo *sinfo;
-  SeedObject *extension;
+  SeedValue extension;
 
   sinfo = (SeedInfo *) g_hash_table_lookup (sloader->loaded_plugins, info);
   if (!sinfo)
@@ -145,35 +149,83 @@ peas_plugin_loader_seed_provides_extension  (PeasPluginLoader *loader,
   extension = seed_object_get_property (sinfo->context,
                                         sinfo->extensions,
                                         g_type_name (exten_type));
-  return extension != NULL;
+  return extension && seed_value_is_object (sinfo->context, extension);
 }
 
 static PeasExtension *
-peas_plugin_loader_seed_get_extension (PeasPluginLoader *loader,
-                                       PeasPluginInfo   *info,
-                                       GType             exten_type)
+peas_plugin_loader_seed_create_extension (PeasPluginLoader *loader,
+                                          PeasPluginInfo   *info,
+                                          GType             exten_type,
+                                          guint             n_parameters,
+                                          GParameter       *parameters)
 {
   PeasPluginLoaderSeed *sloader = PEAS_PLUGIN_LOADER_SEED (loader);
   SeedInfo *sinfo;
-  SeedValue extension;
+  SeedValue extension_methods, extension;
+  gchar **property_names;
+  guint i, j;
 
   sinfo = (SeedInfo *) g_hash_table_lookup (sloader->loaded_plugins, info);
   if (!sinfo)
     return NULL;
 
-  extension = seed_object_get_property (sinfo->context,
-                                        sinfo->extensions,
-                                        g_type_name (exten_type));
-  if (!extension)
+  /* FIXME: instantiate new object and pass the parameters */
+  extension_methods = seed_object_get_property (sinfo->context,
+                                                sinfo->extensions,
+                                                g_type_name (exten_type));
+  if (!extension_methods || seed_value_is_undefined (sinfo->context, extension_methods) || seed_value_is_null (sinfo->context, extension_methods))
     return NULL;
 
-  if (!seed_value_is_object (sinfo->context, extension))
+  if (!seed_value_is_object (sinfo->context, extension_methods))
     {
-      g_warning ("Extension %s in plugin %s is not a javascript object.",
+      g_warning ("Extension '%s' in plugin '%s' is not a Seed object",
                  g_type_name (exten_type), peas_plugin_info_get_module_name (info));
       return NULL;
     }
 
+  /* Copy the original extension_methods object to a new specific object. */
+  extension = seed_make_object (sinfo->context, NULL, NULL);
+  property_names = seed_object_copy_property_names (sinfo->context,
+                                                    extension_methods);
+  for (i = 0; property_names[i] != NULL; i++)
+    {
+      SeedValue value;
+
+      value = seed_object_get_property (sinfo->context,
+                                        extension_methods,
+                                        property_names[i]);
+      seed_object_set_property (sinfo->context,
+                                extension,
+                                property_names[i],
+                                value);
+    }
+
+  /* Set the properties as well */
+  for (i = 0; i < n_parameters; i++)
+    {
+      gchar *key;
+      SeedValue value;
+
+      /* We want to normalize the property names to have a '_' instead of the
+       * conventional '-', to make them accessible through this.property_name */
+      key = g_strdup (parameters[i].name);
+      for (j = 0; key[j] != '\0'; j++)
+        if (key[j] == '-')
+          key[j] = '_';
+
+      value = seed_value_from_gvalue (sinfo->context,
+                                      &parameters[i].value,
+                                      NULL);
+      seed_object_set_property (sinfo->context,
+                                extension,
+                                key,
+                                value);
+
+      g_free (key);
+    }
+
+  g_strfreev (property_names);
+
   return peas_extension_seed_new (exten_type, sinfo->context, extension);
 }
 
@@ -220,22 +272,14 @@ peas_plugin_loader_seed_class_init (PeasPluginLoaderSeedClass *klass)
   loader_class->add_module_directory = peas_plugin_loader_seed_add_module_directory;
   loader_class->load = peas_plugin_loader_seed_load;
   loader_class->provides_extension = peas_plugin_loader_seed_provides_extension;
-  loader_class->get_extension = peas_plugin_loader_seed_get_extension;
+  loader_class->create_extension = peas_plugin_loader_seed_create_extension;
   loader_class->unload = peas_plugin_loader_seed_unload;
   loader_class->garbage_collect = peas_plugin_loader_seed_garbage_collect;
 }
 
-static void
-peas_plugin_loader_seed_class_finalize (PeasPluginLoaderSeedClass *klass)
-{
-}
-
 G_MODULE_EXPORT void
 peas_register_types (PeasObjectModule *module)
 {
-  peas_plugin_loader_seed_register_type (G_TYPE_MODULE (module));
-  peas_extension_seed_register (G_TYPE_MODULE (module));
-
   peas_object_module_register_extension_type (module,
                                               PEAS_TYPE_PLUGIN_LOADER,
                                               PEAS_TYPE_PLUGIN_LOADER_SEED);
